import pickle

import pandas as pd
from lightgbm import LGBMClassifier

from player_performance_ratings.ratings.opponent_adjusted_rating import TeamRatingGenerator
from player_performance_ratings.ratings.opponent_adjusted_rating.performance_predictor import \
    RatingDifferencePerformancePredictor, RatingMeanPerformancePredictor

from player_performance_ratings.transformation import LagTransformer, RollingMeanTransformer

from player_performance_ratings.ratings import ColumnWeight
from player_performance_ratings.tuner.predictor_tuner import PredictorTuner
from player_performance_ratings.tuner.rating_generator_tuner import OpponentAdjustedRatingGeneratorTuner

from player_performance_ratings.scorer import LogLossScorer

from player_performance_ratings import ColumnNames

from player_performance_ratings.tuner.utils import ParameterSearchRange, get_default_team_rating_search_range, \
    get_default_lgbm_classifier_search_range_by_learning_rate
from venn_abers import VennAbersCalibrator

from player_performance_ratings.consts import PredictColumnNames

from player_performance_ratings.predictor.estimators.sklearn_models import SkLearnWrapper

from player_performance_ratings.ratings.opponent_adjusted_rating import OpponentAdjustedRatingGenerator
from player_performance_ratings.tuner import MatchPredictorTuner
from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory

column_names = ColumnNames(
    team_id='team_id',
    match_id='game_id',
    start_date="start_date",
    player_id="player_id",
    performance="plus_minus_per_minute",
    participation_weight="participation_weight",
    projected_participation_weight="projected_participation_weight",

)

df = pd.read_pickle("data/game_player_full_with_minutes_prediction.pickle")

# df = df[df['game_id'].isin(gm['game_id'].unique().tolist())]


df[PredictColumnNames.TARGET] = df['won']
df = df.sort_values(by=[column_names.start_date, column_names.match_id, column_names.team_id, column_names.player_id])
df = df[df['game_minutes'] > 46]
df = df.rename(columns={'__target_prediction': 'minutes_prediction'})
df['plus_minus_per_minute'] = df['plus_minus'] / df['game_minutes']
df['participation_weight'] = df['minutes'] / df['game_minutes']
df['projected_participation_weight'] = df['minutes_prediction'] / df['game_minutes'].mean()
df['score_difference'] = df['score'] - df['score_opponent']

df = (
    df.assign(team_count=df.groupby('game_id')['team_id'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
    .drop(columns=['team_count'])
)
df = df.drop_duplicates(subset=['game_id', 'player_id'])

start_rating_search_range = [
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.04,
        high=.4,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='int',
        low=50,
        high=400,
    ),
]
df.loc[df['minutes'] > 0, 'plus_minus_per_minute'] = df['plus_minus'] / df['minutes']
df.loc[df['minutes'] == 0, 'plus_minus_per_minute'] = 0

estimator = SkLearnWrapper(
    VennAbersCalibrator(
        estimator=LGBMClassifier(max_depth=2, learning_rate=0.1, n_estimators=200, verbose=-100, reg_alpha=1),
        inductive=True, cal_size=0.2, random_state=101))

performance_predictor = RatingDifferencePerformancePredictor(
    rating_diff_team_from_entity_coef=0.00425,
)


rating_generator = OpponentAdjustedRatingGenerator(column_names=column_names, team_rating_generator=TeamRatingGenerator(
    performance_predictor=performance_predictor))

column_weights = [ColumnWeight(name='plus_minus', weight=1)]


post_rating_transformers = [
    LagTransformer(
        features=["score_difference"],
        lag_length=5,
        granularity=[column_names.player_id],
        column_names=column_names,
        days_between_lags=[1, 2, 3, 4, 5],
    ),
    LagTransformer(
        features=[],
        lag_length=4,
        granularity=[column_names.player_id],
        column_names=column_names,
        prefix="future_lag",
        future_lag=True,
        days_between_lags=[1, 2, 3, 4]
    ),

    RollingMeanTransformer(
        features=["score_difference"],
        window=800,
        min_periods=300,
        granularity=["location"],
        column_names=column_names,
    )
]

match_predictor_factory = MatchPredictorFactory(
    post_rating_transformers=post_rating_transformers,
    use_auto_create_performance_calculator=True,
    rating_generators=rating_generator,
    other_categorical_features=["location"],
    estimator=estimator,
    date_column_name=column_names.start_date,
    column_weights=column_weights,
    group_predictor_by_game_team=True,
    team_id_column_name=column_names.team_id,
    match_id_column_name=column_names.match_id,
    train_split_date="2022-12-11",
)

rating_generator_tuner = OpponentAdjustedRatingGeneratorTuner(
    team_rating_search_ranges=get_default_team_rating_search_range(),
    start_rating_search_ranges=start_rating_search_range,
    team_rating_n_trials=3,
    start_rating_n_trials=8,
)
predictor_tuner = PredictorTuner(
    default_params={'learning_rate': 0.03},
    search_ranges=get_default_lgbm_classifier_search_range_by_learning_rate(learning_rate=0.03),
    n_trials=65,
    date_column_name=column_names.start_date,
    estimator_subclass_level=2,
)

tuner = MatchPredictorTuner(
    match_predictor_factory=match_predictor_factory,
    fit_best=True,
    scorer=LogLossScorer(pred_column=match_predictor_factory.predictor.pred_column),
    rating_generator_tuners=rating_generator_tuner,
    predictor_tuner=predictor_tuner,
)
best_match_predictor = tuner.tune(df=df)
df_with_minutes_prediction = best_match_predictor.generate_historical(df=df)

pickle.dump(best_match_predictor, open("models/nba_game_winner", 'wb'))
df_with_minutes_prediction.to_pickle("data/game_player_predictions.pickle")



import logging
import warnings
import pandas as pd
from lightgbm import LGBMClassifier
from pandas.errors import SettingWithCopyWarning
from player_performance_ratings.transformation.base_transformer import BaseTransformer


warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

from typing import Optional, Any

from sklearn.linear_model import LogisticRegression

from player_performance_ratings.consts import PredictColumnNames

from player_performance_ratings.data_structures import ColumnNames
from player_performance_ratings.predictor.estimators.base_estimator import BaseMLWrapper


class GameTeamPredictor(BaseMLWrapper):

    def __init__(self,
                 game_id_colum: str,
                 team_id_column: str,
                 features: list[str],
                 weight_column: Optional[str] = None,
                 target: Optional[str] = PredictColumnNames.TARGET,
                 estimator: Optional = None,
                 multiclassifier: bool = False,
                 pred_column: Optional[str] = None,
                 categorical_transformers: Optional[list[BaseTransformer]] = None

                 ):
        """
        Wrapper for sklearn models that predicts game results.

        The GameTeam Predictor is intended to transform predictions from a lower granularity into a GameTeam level.
        So if input data is at game-player, data is converted to game_team before being trained
        Similar concept if it is at a granularity below game level.

        The weight_column makes it possible to weight certain rows more than others.
        For instance, if data is on game-player level and the participation rate of players is not equal -->
         setting participation_weight equal to weight_column will the feature values of the players with high participation_rates have higher weight before aggregating to game_team.


        :param game_id_colum:
        :param team_id_column:
        :param features:
        :param weight_column:
        :param target:
        :param estimator:
        :param multiclassifier:
        :param pred_column:
        """

        self.weight_column = weight_column
        self.game_id_colum = game_id_colum
        self.team_id_column = team_id_column
        self._target = target

        self.multiclassifier = multiclassifier
        super().__init__(target=self._target, features=features, pred_column=pred_column,
                         estimator=estimator or LogisticRegression(), categorical_transformers=categorical_transformers)

    def train(self, df: pd.DataFrame) -> None:
        df = self.fit_transform_categorical_transformers(df=df)
        if len(df[self._target].unique()) > 2 and hasattr(self.estimator, "predict_proba"):
            logging.warning("target has more than 2 unique values, multiclassifier has therefore been set to True")
            self.multiclassifier = True

        if self._target not in df.columns:
            raise ValueError(f"target {self._target} not in df")
        grouped = self._create_grouped(df)
        self.estimator.fit(grouped[self.estimator_features], grouped[self._target])

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Adds prediction to df

        :param df:
        :return: Input df with prediction column
        """
        df = self.transform_categorical_transformers(df=df)
        grouped = self._create_grouped(df)

        if self.multiclassifier:
            grouped[self._pred_column] = self.estimator.predict_proba(grouped[self.estimator_features]).tolist()
        elif not hasattr(self.estimator, "predict_proba"):
            grouped[self._pred_column] = self.estimator.predict(grouped[self.estimator_features])
        else:
            grouped[self._pred_column] = self.estimator.predict_proba(grouped[self.estimator_features])[:, 1]

        if self.pred_column in df.columns:
            df = df.drop(columns=[self.pred_column])

        df = df.merge(grouped[[self.game_id_colum, self.team_id_column] + [self._pred_column]],
                      on=[self.game_id_colum, self.team_id_column])

        return df

    def _create_grouped(self, df: pd.DataFrame) -> pd.DataFrame:

        numeric_features = [feature for feature in self.estimator_features if
                            feature not in self.estimator_categorical_features]

        if self._target in df.columns:
            if df[self._target].dtype == 'object':
                df.loc[:, self._target] = df[self._target].astype('int')

        if self.weight_column:
            for feature in numeric_features:
                df = df.assign(**{feature: df[self.weight_column] * df[feature]})

        if self.weight_column:
            grouped = df.groupby([self.game_id_colum, self.team_id_column]).agg({
                **{feature: 'sum' for feature in numeric_features},
                self._target: 'mean',
                self.weight_column: 'sum',
            }).reset_index()

            for feature in numeric_features:
                grouped[feature] = grouped[feature] / grouped[self.weight_column]

            grouped.drop(columns=[self.weight_column], inplace=True)

        else:
            if self._target in df.columns:
                grouped = df.groupby([self.game_id_colum, self.team_id_column]).agg({
                    **{feature: 'mean' for feature in numeric_features},
                    self._target: 'mean',
                }).reset_index()
            else:
                grouped = df.groupby([self.game_id_colum, self.team_id_column]).agg({
                    **{feature: 'mean' for feature in numeric_features}
                }).reset_index()

        if self._target in df.columns:
            grouped[self._target] = grouped[self._target].astype('int')

        grouped = grouped.merge(df[[self.game_id_colum, self.team_id_column, *self.estimator_categorical_features]].drop_duplicates(
                                    subset=[self.game_id_colum, self.team_id_column]),
                                on=[self.game_id_colum, self.team_id_column], how='inner')
        return grouped


class Predictor(BaseMLWrapper):

    def __init__(self,
                 features: list[str],
                 target: Optional[str] = PredictColumnNames.TARGET,
                 estimator: Optional = None,
                 multiclassifier: bool = False,
                 pred_column: Optional[str] = None,
                 column_names: Optional[ColumnNames] = None,
                 categorical_transformers: Optional[list[BaseTransformer]] = None
                 ):
        self._target = target
        self.multiclassifier = multiclassifier
        self.column_names = column_names

        if estimator is None:
            logging.warning(
                "model is not set. Will use LGBMClassifier(max_depth=2, n_estimators=400, learning_rate=0.05)")

        super().__init__(target=self._target, features=features, pred_column=pred_column,
                         estimator=estimator or LGBMClassifier(max_depth=2, n_estimators=300, learning_rate=0.05, verbose=-100),
                         categorical_transformers=categorical_transformers)

    def train(self, df: pd.DataFrame) -> None:
        df = self.fit_transform_categorical_transformers(df=df)
        if self.multiclassifier is False and len(df[self._target].unique()) > 2 and hasattr(self.estimator,
                                                                                            "predict_proba"):
            logging.warning("target has more than 2 unique values, multiclassifier has therefore been set to True")
            self.multiclassifier = True
            if len(df[self._target].unique()) > 50:
                logging.warning(
                    f"target has {len(df[self._target].unique())} unique values. This may machine-learning model to not function properly."
                    f" It is recommended to limit max and min values to ensure less than 50 unique targets")

        if hasattr(self.estimator, "predict_proba"):
            df = df.assign(**{self._target: df[self._target].astype('int')})

        self.estimator.fit(df[self.estimator_features], df[self._target])

    def add_prediction(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.transform_categorical_transformers(df=df)
        df = df.copy()

        if self.multiclassifier:
            df[self._pred_column] = self.estimator.predict_proba(df[self.estimator_features]).tolist()
            if len(set(df[self.pred_column].iloc[0])) == 2:
                raise ValueError(
                    "Too many unique values in relation to rows in the training dataset causes multiclassifier to not train properly")

        elif not hasattr(self.estimator, "predict_proba"):
            df[self._pred_column] = self.estimator.predict(df[self.estimator_features])
        else:
            df[self._pred_column] = self.estimator.predict_proba(df[self.estimator_features])[:, 1]
        return df



import pandas as pd

df = pd.read_pickle("game_player.pickle")
gt = pd.read_pickle("game_team.pickle")
game = pd.read_pickle("game.pickle")
game = game.rename(columns={'minutes': 'game_minutes'})
df = df.merge(game[['game_id', 'game_minutes', 'start_date']], on='game_id', how='left')
df = df.merge(gt[['game_id', 'team_id', "team_id_opponent", 'score', 'score_opponent', 'won', "location"]],
              on=['game_id', 'team_id'], how='inner')
df.loc[df['won'] == True, 'won'] = 1
df.loc[df['won'] == False, 'won'] = 0

# randomly select 10% of game_ids

df = df[["team_id", "start_date", "game_id", "player_id", 'player_name', 'start_position', "team_id_opponent", "points",
         "game_minutes",
         "minutes", "won", "plus_minus", "location", "score", "score_opponent"]]
#df = df[df['start_date'].between('2022-10-17', '2023-02-01')]

df.to_pickle("game_player_full.pickle")



import logging
from dataclasses import dataclass

import numpy as np
from typing import Dict, Any, List, Optional

from player_performance_ratings.data_structures import MatchPlayer, PreMatchPlayerRating, PlayerRatingChange

DEFAULT_START_RATING = 1000


@dataclass
class LeagueEntityRatings:
    league: str
    entity_ratings: List[float]


class StartRatingGenerator():

    def __init__(self,
                 league_ratings: Optional[dict[str, float]] = None,
                 league_quantile: float = 0.2,
                 min_count_for_percentiles: int = 100,
                 team_rating_subtract: float = 80,
                 team_weight: float = 0,
                 max_days_ago_league_entities: int = 120,
                 min_match_count_team_rating: int = 2,
                 harcoded_start_rating: Optional[float] = None,
                 ):

        self.league_ratings = league_ratings or {}
        self.league_quantile = league_quantile
        self.min_count_for_percentiles = min_count_for_percentiles
        self.team_rating_subtract = team_rating_subtract
        self.team_weight = team_weight
        self.max_days_ago_league_entities = max_days_ago_league_entities
        self.min_match_count_team_rating = min_match_count_team_rating
        self.harcoded_start_rating = harcoded_start_rating
        if self.harcoded_start_rating is not None:
            logging.warning(
                f"Hardcoded ratings are used. This will usually result in worse accuracy."
                f"Although can come with computational benefits if there are many different players in the dataset")

        self._league_to_last_day_number: Dict[str, List[Any]] = {}
        self._league_to_entity_ids: Dict[str, List[str]] = {}
        self._league_player_ratings: dict[str, list] = {}
        self._entity_to_league: Dict[str, str] = {}

    def generate_rating_value(self,
                              day_number: int,
                              match_entity: MatchPlayer,
                              team_pre_match_player_ratings: list[PreMatchPlayerRating],
                              ) -> float:

        if self.harcoded_start_rating is not None:
            return self.harcoded_start_rating

        league = match_entity.league
        if league not in self.league_ratings:
            self.league_ratings[league] = DEFAULT_START_RATING

        if league not in self._league_to_entity_ids:
            self._league_to_entity_ids[league] = []
            self._league_to_last_day_number[league] = []

        if league not in self._league_player_ratings:
            self._league_player_ratings[league] = []

        existing_team_rating = None
        if self.team_weight > 0:
            tot_player_game_count = sum([p.games_played for p in team_pre_match_player_ratings])
            if tot_player_game_count >= self.min_match_count_team_rating:
                sum_team_rating = sum(player.rating_value * player.match_performance.participation_weight for player in
                                      team_pre_match_player_ratings)
                sum_participation_weight = sum(
                    player.match_performance.participation_weight for player in team_pre_match_player_ratings)

                existing_team_rating = sum_team_rating / sum_participation_weight if sum_participation_weight > 0 else 0

        if existing_team_rating is None:
            team_weight = 0
            adjusted_team_start_rating = 0
        else:
            team_weight = self.team_weight

            adjusted_team_start_rating = existing_team_rating - self.team_rating_subtract
        start_rating_league = self._calculate_start_rating_value(
            match_day_number=day_number,
            league=league,
        )
        return start_rating_league * (1 - team_weight) + team_weight * adjusted_team_start_rating

    def _calculate_start_rating_value(self,
                                      match_day_number: int,
                                      league: str,
                                      ) -> float:
        new_entity_ratings = self._get_new_entities_ratings(
            match_day_number=match_day_number,
            league=league,
        )
        region_entity_count = len(new_entity_ratings)
        if region_entity_count < self.min_count_for_percentiles:
            return self.league_ratings[league]
        else:
            return self._start_rating_value_for_above_threshold(new_entity_ratings)

    def _get_new_entities_ratings(self,
                                  match_day_number: int,
                                  league: str,
                                  ) -> List[float]:
        entity_ratings: List[float] = []

        for index, last_day_number in enumerate(self._league_to_last_day_number[league]):
            days_ago = match_day_number - last_day_number

            if days_ago <= self.max_days_ago_league_entities:
                entity_ratings.append(self._league_player_ratings[
                                          league][index])

        return entity_ratings

    def _start_rating_value_for_above_threshold(self, entity_ratings: List) -> float:
        percentile = np.percentile(entity_ratings, self.league_quantile * 100)
        return percentile

    def update_league_ratings(self, rating_change: PlayerRatingChange):
        league = rating_change.league
        id = rating_change.id
        day_number = rating_change.day_number
        rating_value = rating_change.pre_match_rating_value + rating_change.rating_change_value

        league_data = self._league_player_ratings.setdefault(league, [])
        league_entity_ids = self._league_to_entity_ids.setdefault(league, [])
        league_last_day_numbers = self._league_to_last_day_number.setdefault(league, [])

        if id not in league_entity_ids:
            league_data.append(rating_value)
            league_entity_ids.append(id)
            league_last_day_numbers.append(day_number)
            self._entity_to_league[id] = league
        else:

            index = league_entity_ids.index(id)
            league_last_day_numbers[index] = day_number
            league_data[index] = rating_value

        current_entity_league = self._entity_to_league.get(id, league)
        if league != current_entity_league:
            entity_index = self._league_to_entity_ids[current_entity_league].index(id)

            for data_structure in (self._league_player_ratings[current_entity_league],
                                   self._league_to_last_day_number[current_entity_league],
                                   self._league_to_entity_ids[current_entity_league]):
                del data_structure[entity_index]

            self._entity_to_league[id] = league

    @property
    def league_player_ratings(self) -> dict[str, list]:
        return self._league_player_ratings


import pandas as pd


class LineupMatchupPerformance():

    def __init__(self, lineup_matchup: pd.DataFrame):
        self.lineup_matchup = lineup_matchup

    def generate(self, df: pd.DataFrame):
        q = self.lineup_matchup.explode('LINEUP_OFFENSE')
        u = 2


if __name__ == '__main__':
    game = pd.read_pickle("data/game.pickle")
    game = game.rename(columns={'MINUTES': 'GAME_MINUTES'})
    game_team = pd.read_pickle("data/game_team.pickle")
    game_player = pd.read_pickle("data/game_player.pickle")

    df = game_player.merge(game_team, on=['GAME_ID', 'TEAM_ID'], how='left')
    df = df.merge(game, on=['GAME_ID'], how='left')

    df.loc[df['WON']==True, 'WON'] = 1
    df.loc[df['WON'] == False, 'WON'] = 0

    lineup_matchup = pd.read_pickle("data/lineup_matchup.pickle")
    l = LineupMatchupPerformance(lineup_matchup)
    l.generate(df)



import pandas as pd
from lightgbm import LGBMClassifier


from player_performance_ratings.predictor.estimators import Predictor

from player_performance_ratings import ColumnNames, PredictColumnNames

from player_performance_ratings.predictor import MatchPredictor
from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.ratings.time_weight_ratings import BayesianTimeWeightedRating
from player_performance_ratings.scorer.score import OrdinalLossScorer

column_names_kpm = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='kills_per_minute',
    league='league',
    position='position',
)

column_names_kills = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='kills',
    league='league',
    position='position'
)

df = pd.read_parquet("data/subsample_lol_data")
df = df.sort_values(by=['date', 'gameid', 'teamname', "playername"])
df['__target'] = df['kills']
df["__target"] = df["__target"].clip(0, 8)

df['kills_per_minute'] = df['kills'] / df['gamelength'] * 60

df = (
    df.loc[lambda x: x.position != 'team']
    .assign(team_count=df.groupby('gameid')['teamname'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)

time_weighed_rating_kills_per_minute = BayesianTimeWeightedRating(column_names=column_names_kpm)
time_weighed_rating_kills = BayesianTimeWeightedRating(column_names=column_names_kills)

match_predictor = MatchPredictor(
    rating_generators=[time_weighed_rating_kills_per_minute, time_weighed_rating_kills],
    use_auto_create_performance_calculator=True,
    column_weights=[[ColumnWeight(name='kills_per_minute', weight=1)], [ColumnWeight(name='kills', weight=1)]],
    other_categorical_features=["position"],
    predictor=Predictor(
        estimator=LGBMClassifier(verbose=-100),
        features=[
            RatingColumnNames.TIME_WEIGHTED_RATING + "0",
            RatingColumnNames.TIME_WEIGHTED_RATING + "1",
            RatingColumnNames.TIME_WEIGHTED_RATING_LIKELIHOOD_RATIO + "0",
            RatingColumnNames.TIME_WEIGHTED_RATING_LIKELIHOOD_RATIO + "1",
            RatingColumnNames.TIME_WEIGHTED_RATING_EVIDENCE + "0",
            RatingColumnNames.TIME_WEIGHTED_RATING_EVIDENCE + "1",
            "position"
        ],
        target=PredictColumnNames.TARGET,
    )
)

df_predictions = match_predictor.generate_historical(df=df)

for idx, kills in enumerate(match_predictor.predictor.estimator.classes_):
    print(df_predictions.iloc[500]['playername'], kills,
          df_predictions.iloc[500][match_predictor.predictor.pred_column][idx])

print(match_predictor.predictor.estimator.feature_importances_)

scorer = OrdinalLossScorer(
    pred_column=match_predictor.predictor.pred_column,
)

print(scorer.score(df_predictions, classes_=match_predictor.predictor.estimator.classes_))



import copy

from typing import Optional, Union

import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial

from player_performance_ratings.data_structures import Match
from player_performance_ratings.ratings import ColumnWeight, PerformancesGenerator
from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.scorer.score import BaseScorer
from player_performance_ratings.transformation.base_transformer import BaseTransformer

from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory

from player_performance_ratings.tuner.utils import add_params_from_search_range, ParameterSearchRange

RC = RatingColumnNames


class PerformancesGeneratorTuner:

    def __init__(self,
                 performances_weight_search_ranges: dict[str, list[ParameterSearchRange]],
                 pre_transformations: Optional[list[BaseTransformer]] = None,
                 feature_names: Optional[Union[list[str], list[list[str]]]] = None,
                 lower_is_better_features: Optional[list[str]] = None,
                 n_trials: int = 30,
                 ):

        self.performances_weight_search_ranges = performances_weight_search_ranges
        self.pre_transformations = pre_transformations
        self.feature_names = feature_names
        self.lower_is_better_features = lower_is_better_features

        self.n_trials = n_trials
        self._scores = []

    def tune(self, df: pd.DataFrame,
             scorer: BaseScorer,
             match_predictor_factory: MatchPredictorFactory,
             matches: Optional[list[list[Match]]] = None,
             ) -> PerformancesGenerator:

        df = df.copy()

        if not matches:
            matches = None

        column_names = [r.column_names for r in match_predictor_factory.rating_generators]


        match_predictor_factory = copy.deepcopy(match_predictor_factory)

        def objective(trial: BaseTrial,
                      df: pd.DataFrame,
                      match_predictor_factory: MatchPredictorFactory,
                      scorer: BaseScorer,
                      ) -> float:


            column_weights = []
            for performance_name, search_range in self.performances_weight_search_ranges.items():
                raw_params = {}
                renamed_search = copy.deepcopy(search_range)
                for sr in renamed_search:
                    sr.name = f"{performance_name}__{sr.name}"
                add_params_from_search_range(trial=trial, params=raw_params, parameter_search_range=renamed_search)
                column_weights.append(
                    self._create_column_weights(params=raw_params, remove_string=f"{performance_name}__"))

            col_names = [r.column_names for r in match_predictor_factory.rating_generators]
            performances_generator = PerformancesGenerator(
                column_names=col_names,
                column_weights=column_weights,
                pre_transformations=self.pre_transformations
            )
            match_predictor = match_predictor_factory.create(performances_generator=performances_generator)

            df_with_prediction = match_predictor.generate_historical(df=df, matches=matches, store_ratings=False)
            test_df = df_with_prediction[
                df_with_prediction[match_predictor.date_column_name] > match_predictor.train_split_date]
            return scorer.score(test_df, classes_=match_predictor.predictor.classes_)

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(direction=direction, study_name=study_name, sampler=sampler)
        callbacks = []
        study.optimize(
            lambda trial: objective(trial, df=df,
                                    match_predictor_factory=match_predictor_factory, scorer=scorer),
            n_trials=self.n_trials, callbacks=callbacks)

        best_params = study.best_params
        best_column_weights = self._select_best_column_weights(all_params=best_params)


        return PerformancesGenerator(column_weights=best_column_weights, column_names=column_names,
                                     pre_transformations=self.pre_transformations)

    def _create_column_weights(self, params: dict, remove_string: str) -> list[ColumnWeight]:

        sum_weights = sum([v for _, v in params.items()])
        column_weights = []
        for name, weight in params.items():
            column_weights.append(ColumnWeight(name=name.replace(remove_string, ""), weight=weight / sum_weights))

        return column_weights

    def _select_best_column_weights(self, all_params: dict) -> list[list[ColumnWeight]]:
        best_column_weights = []
        for performance_name in self.performances_weight_search_ranges:
            column_weights = []
            sum_weights = 0
            for param in all_params:

                if f"{performance_name}__" in param:
                    sum_weights += all_params[param]

            for param in all_params:
                if f"{performance_name}__" in param:
                    column_weights.append(ColumnWeight(name=param.replace(f"{performance_name}__", ""), weight=all_params[param] / sum_weights))

            best_column_weights.append(column_weights)

        return best_column_weights



import copy
from typing import Optional, List, Union

import pendulum

from player_performance_ratings.predictor.match_predictor import create_predictor
from player_performance_ratings.ratings import PerformancesGenerator, ColumnWeight

from player_performance_ratings.predictor import MatchPredictor
from player_performance_ratings.predictor.estimators.base_estimator import BaseMLWrapper
from player_performance_ratings.ratings.rating_generator import RatingGenerator
from player_performance_ratings.transformation.base_transformer import BaseTransformer, BasePostTransformer
from player_performance_ratings.transformation.factory import auto_create_performance_generator


class MatchPredictorFactory():

    def __init__(self,
                 date_column_name: Optional[str],
                 rating_generators: Optional[Union[RatingGenerator, list[RatingGenerator]]] = None,
                 performances_generator: Optional[PerformancesGenerator] = None,
                 post_rating_transformers: Optional[List[BasePostTransformer]] = None,
                 post_prediction_transformers: Optional[List[BasePostTransformer]] = None,
                 predictor: BaseMLWrapper = None,
                 estimator: Optional = None,
                 other_features: Optional[list[str]] = None,
                 other_categorical_features: Optional[list[str]] = None,
                 group_predictor_by_game_team: bool = False,
                 train_split_date: Optional[pendulum.datetime] = None,
                 match_id_column_name: Optional[str] = None,
                 team_id_column_name: Optional[str] = None,
                 use_auto_create_performance_calculator: bool = False,
                 column_weights: Optional[Union[List[List[ColumnWeight]], list[ColumnWeight]]] = None,
                 ):


        if use_auto_create_performance_calculator and not column_weights:
            raise ValueError("If auto pre transformers are used, column weights must be specified")

        self.rating_generators = rating_generators or []
        if isinstance(self.rating_generators, RatingGenerator):
            self.rating_generators = [self.rating_generators]

        self.post_rating_transformers = post_rating_transformers or []
        self.post_prediction_transformers = post_prediction_transformers or []

        self.predictor = predictor
        self.estimator = estimator
        self.date_column_name = date_column_name
        self.other_features = other_features or []
        self.other_categorical_features = other_categorical_features or []
        self.group_predictor_by_game_team = group_predictor_by_game_team
        self.match_id_column_name = match_id_column_name
        self.team_id_column_name = team_id_column_name

        self.train_split_date = train_split_date
        self.use_auto_create_performance_calculator = use_auto_create_performance_calculator
        self.performances_generator = performances_generator
        self.column_weights = column_weights if isinstance(column_weights, list) else [column_weights] if column_weights else None

        if self.predictor is None:
            self.predictor = create_predictor(
                rating_generators=self.rating_generators,
                other_features=self.other_features,
                other_categorical_features=self.other_categorical_features,
                post_rating_transformers=self.post_rating_transformers,
                estimator=estimator,
                group_predictor_by_game_team=self.group_predictor_by_game_team,
                match_id_column_name=self.match_id_column_name,
                team_id_column_name=self.team_id_column_name
            )

        if self.use_auto_create_performance_calculator:
            if not self.rating_generators:
                raise ValueError("If auto pre transformers are used, rating generators must be specified")
            column_names = [rating_generator.column_names for rating_generator in self.rating_generators]
            self.performances_generator = auto_create_performance_generator(column_weights=self.column_weights,
                                                                                 column_names=column_names)

    def create(self,
               performances_generator: Optional[PerformancesGenerator] = None,
               rating_generators: Optional[list[RatingGenerator]] = None,
               post_rating_transformers: Optional[List[BaseTransformer]] = None,
               predictor: Optional[BaseMLWrapper] = None,
               ) -> MatchPredictor:

        rating_generators = rating_generators if rating_generators is not None else self.rating_generators
        performances_generator = performances_generator if performances_generator is not None else self.performances_generator
        post_rating_transformers = post_rating_transformers if post_rating_transformers is not None else self.post_rating_transformers
        predictor = predictor if predictor is not None else self.predictor

        return MatchPredictor(
            rating_generators=rating_generators,
            performances_generator=performances_generator,
            post_rating_transformers=copy.deepcopy(post_rating_transformers),
            post_prediction_transformers=copy.deepcopy(self.post_prediction_transformers),
            predictor=predictor,
            train_split_date=self.train_split_date,
            match_id_column_name=self.match_id_column_name,
            date_column_name=self.date_column_name,
        )



from dataclasses import dataclass
from typing import Optional, Union

import pandas as pd
from player_performance_ratings import ColumnNames

from player_performance_ratings.transformation.base_transformer import BaseTransformer

@dataclass
class ColumnWeight:
    name: str
    weight: float
    lower_is_better: bool = False

    def __post_init__(self):
        if self.weight < 0:
            raise ValueError("Weight must be positive")
        if self.weight > 1:
            raise ValueError("Weight must be less than 1")


class PerformancesGenerator():

    def __init__(self,
                 column_weights: Union[list[list[ColumnWeight]], list[ColumnWeight]],
                 column_names: Union[list[ColumnNames], ColumnNames],
                 pre_transformations: Optional[list[BaseTransformer]] = None,
                 ):
        self.column_names = column_names if isinstance(column_names, list) else [column_names]
        self.pre_transformations = pre_transformations
        self.column_weights = column_weights if isinstance(column_weights[0], list) else [column_weights]

    def generate(self, df):
        if self.pre_transformations:
            for pre_transformation in self.pre_transformations:
                df = pre_transformation.fit_transform(df)

        for idx, col_name in enumerate(self.column_names):
            df[col_name.performance] = self._weight_columns(df=df, col_name=col_name, col_weights=self.column_weights[idx])
        return df

    def _weight_columns(self, df: pd.DataFrame, col_name: ColumnNames, col_weights: list[ColumnWeight]) -> pd.DataFrame:
        df = df.copy()
        df[f"__{col_name.performance}"] = 0

        df['sum_cols_weights'] = 0
        for column_weight in col_weights:
            df[f'weight__{column_weight.name}'] = column_weight.weight
            df.loc[df[column_weight.name].isna(), f'weight__{column_weight.name}'] = 0
            df.loc[df[column_weight.name].isna(), column_weight.name] = 0
            df['sum_cols_weights'] = df['sum_cols_weights'] + df[f'weight__{column_weight.name}']

        drop_cols = ['sum_cols_weights', f"__{col_name.performance}"]
        for column_weight in col_weights:
            df[f'weight__{column_weight.name}'] / df['sum_cols_weights']
            drop_cols.append(f'weight__{column_weight.name}')

        for column_weight in col_weights:

            if column_weight.lower_is_better:
                df[f"__{col_name.performance}"] += df[f'weight__{column_weight.name}'] * (
                        1 - df[column_weight.name])
            else:
                df[f"__{col_name.performance}"] += df[f'weight__{column_weight.name}'] * df[column_weight.name]

        return  df[f"__{col_name.performance}"]


    @property
    def features_out(self) -> list[str]:
        return [c.performance for c in self.column_names]




from .performance_predictor import RatingDifferencePerformancePredictor, RatingMeanPerformancePredictor
from .start_rating_generator import StartRatingGenerator
from .team_rating_generator import TeamRatingGenerator
from .opponent_adjusted_rating_generator import OpponentAdjustedRatingGenerator


import copy
import logging
from typing import Optional, Union

import pandas as pd

from player_performance_ratings.predictor.match_predictor import MatchPredictor
from player_performance_ratings.ratings import PerformancesGenerator
from player_performance_ratings.ratings.match_generator import convert_df_to_matches
from player_performance_ratings.scorer import BaseScorer

from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory
from player_performance_ratings.tuner.predictor_tuner import PredictorTuner
from player_performance_ratings.tuner.rating_generator_tuner import RatingGeneratorTuner
from player_performance_ratings.tuner.performances_generator_tuner import PerformancesGeneratorTuner

logging.basicConfig(level=logging.INFO)


class MatchPredictorTuner():
    """
    Given a search range, it will identify the best combination of pre_transformers, rating_generators and post_transformers.
    Using optuna, it will tune the hyperparameters of each transformer.
    """

    def __init__(self,
                 scorer: BaseScorer,
                 match_predictor_factory: MatchPredictorFactory,
                 performances_generator_tuner: Optional[PerformancesGeneratorTuner] = None,
                 rating_generator_tuners: Optional[Union[list[RatingGeneratorTuner], RatingGeneratorTuner]] = None,
                 predictor_tuner: Optional[PredictorTuner] = None,
                 fit_best: bool = True
                 ):

        """
        :param scorer:
            The class object that scores the predictions
        :param match_predictor_factory:
            The factory that creates the MatchPredictor.
            Contains the parameters to create the MatchPredictor if no parameter-tuning was done.
            Based on hyperparameter-tuning the final way the match-predictor is generated will be based on a combination of the parameters in the factory and the tuned parameters.
        :param performances_generator_tuner:
            The tuner that tunes the hyperparameters of the pre_transformers.
            If left none or as [], the pre_transformers in the match_predictor_factory will be used.
        :param rating_generator_tuners:
            The tuner that tunes the hyperparameters of the rating_generators.
            If left none or as [], the rating_generators in the match_predictor_factory will be used.

        :param predictor_tuner:
            The tuner that tunes the hyperparameters of the predictor model.

        :param fit_best: Whether to refit the match_predictor with the entire training use data using best hyperparameters
        """

        self.scorer = scorer
        self.match_predictor_factory = match_predictor_factory
        self.performances_generator_tuner = performances_generator_tuner
        self.rating_generator_tuners = rating_generator_tuners or []
        self.predictor_tuner = predictor_tuner
        if isinstance(self.rating_generator_tuners, RatingGeneratorTuner):
            self.rating_generator_tuners = [self.rating_generator_tuners]
        self.fit_best = fit_best

        if not self.performances_generator_tuner and not self.rating_generator_tuners and not self.predictor_tuner:
            raise ValueError("No tuning has been provided in config")

    def tune(self, df: pd.DataFrame) -> MatchPredictor:


        original_df = df.copy()

        column_names = [rating_generator.column_names for rating_generator in
                        self.match_predictor_factory.rating_generators]

        best_performances_generator: PerformancesGenerator = copy.deepcopy(self.match_predictor_factory.performances_generator)

        if self.performances_generator_tuner:
            logging.info("Tuning PreTransformers")
            best_performances_generator = self.performances_generator_tuner.tune(df=df,
                                                                                 match_predictor_factory=self.match_predictor_factory,
                                                                                 scorer=self.scorer)
        if best_performances_generator:
            df = best_performances_generator.generate(df)

        best_rating_generators = copy.deepcopy(self.match_predictor_factory.rating_generators)

        matches = []
        for col_name in column_names:
            rating_matches = convert_df_to_matches(df=df, column_names=col_name)
            matches.append(rating_matches)

        for rating_idx, rating_generator_tuner in enumerate(self.rating_generator_tuners):
            rating_matches = convert_df_to_matches(df=df, column_names=column_names[rating_idx])
            matches.append(rating_matches)
            tuned_rating_generator = rating_generator_tuner.tune(df=df, matches=matches[rating_idx],
                                                                 rating_idx=rating_idx,
                                                                 scorer=self.scorer,
                                                                 match_predictor_factory=self.match_predictor_factory)
            if best_rating_generators:
                best_rating_generators[rating_idx] = tuned_rating_generator
            else:
                best_rating_generators = [tuned_rating_generator]

            match_ratings = tuned_rating_generator.generate_historical(df=df, matches=matches[rating_idx])
            for rating_feature in tuned_rating_generator.features_out:
                values = match_ratings[rating_feature]

                if len(self.rating_generator_tuners) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature
                df[rating_feature_str] = values


        best_post_transformers = copy.deepcopy(self.match_predictor_factory.post_rating_transformers)

        for post_rating_transformer in best_post_transformers:
            df = post_rating_transformer.fit_transform(df)

        if self.predictor_tuner:
            logging.info("Tuning Predictor")
            best_predictor = self.predictor_tuner.tune(df=df,  scorer=self.scorer,
                                                       match_predictor_factory=self.match_predictor_factory)
        else:
            best_predictor = self.match_predictor_factory.predictor

        best_match_predictor = MatchPredictor(
            rating_generators=best_rating_generators,
            performances_generator=best_performances_generator,
            post_rating_transformers=best_post_transformers,
            post_prediction_transformers=self.match_predictor_factory.post_prediction_transformers,
            predictor=best_predictor)

        if self.fit_best:
            logging.info("Retraining best match predictor with all data")
            best_match_predictor.generate_historical(df=original_df, store_ratings=True)

        return best_match_predictor



from dataclasses import dataclass
from enum import Enum


@dataclass
class InputColumnNames:
    hour = 'hour'
    performance = 'performance'


@dataclass
class RatingColumnNames:
    PLAYER_RATING = 'player_rating'
    PLAYER_RATING_DIFFERENCE_FROM_TEAM_PROJECTED = "player_rating_difference_from_team_projected"
    OPPONENT_RATING_PROJECTED = "opponent_rating_projected"
    TEAM_RATING_PROJECTED = 'team_rating_projected'
    PLAYER_RATING_DIFFERENCE_PROJECTED = 'player_rating_difference_projected'
    RATING_DIFFERENCE_PROJECTED = 'rating_difference_projected'
    RATING_MEAN_PROJECTED = 'rating_mean_projected'
    PLAYER_LEAGUE = "player_league"
    OPPONENT_LEAGUE = "opponent_league"
    MATCH_ID = "match_id"
    TIME_WEIGHTED_RATING = "time_weighted_rating"
    TIME_WEIGHTED_RATING_LIKELIHOOD_RATIO = "time_weighted_rating_likelihood_ratio"
    TIME_WEIGHTED_RATING_EVIDENCE = "time_weighted_rating_evidence"


@dataclass
class HistoricalRatingColumnNames:
    PERFORMANCE = "performance"
    RATING_DIFFERENCE = 'rating_difference'
    PLAYER_RATING_DIFFERENCE = 'player_rating_difference'
    TEAM_RATING = 'team_rating'
    OPPONENT_RATING = "opponent_rating"
    PLAYER_RATING_DIFFERENCE_FROM_TEAM = "player_rating_difference_from_team"
    PLAYER_RATING_CHANGE = "player_rating_change"
    RATING_MEAN = 'rating_mean'
    PLAYER_PREDICTED_PERFORMANCE = "player_predicted_performance"


class PredictedRatingMethod(Enum):
    DEFAULT = 'default'
    MEAN_RATING = "mean_rating"



import copy
import inspect
import logging
from abc import abstractmethod, ABC
from typing import Optional, Match

import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial
from player_performance_ratings.ratings import RatingColumnNames

from player_performance_ratings.ratings.opponent_adjusted_rating import TeamRatingGenerator

from player_performance_ratings.ratings.opponent_adjusted_rating.start_rating_generator import StartRatingGenerator
from player_performance_ratings.ratings.opponent_adjusted_rating import \
    OpponentAdjustedRatingGenerator
from player_performance_ratings.ratings.rating_generator import RatingGenerator
from player_performance_ratings.scorer import BaseScorer

from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory
from player_performance_ratings.tuner.utils import ParameterSearchRange, add_params_from_search_range

DEFAULT_TEAM_SEARCH_RANGES = [
    ParameterSearchRange(
        name='confidence_weight',
        type='uniform',
        low=0.6,
        high=0.95
    ),
    ParameterSearchRange(
        name='confidence_days_ago_multiplier',
        type='uniform',
        low=0.02,
        high=.12,
    ),
    ParameterSearchRange(
        name='confidence_max_days',
        type='uniform',
        low=40,
        high=200,
    ),
    ParameterSearchRange(
        name='confidence_max_sum',
        type='uniform',
        low=20,
        high=180,
    ),
    ParameterSearchRange(
        name='confidence_value_denom',
        type='uniform',
        low=20,
        high=180,
    ),
    ParameterSearchRange(
        name='rating_change_multiplier',
        type='uniform',
        low=25,
        high=140
    ),
    ParameterSearchRange(
        name='min_rating_change_multiplier_ratio',
        type='uniform',
        low=0.05,
        high=0.2,
    ),
]

DEFAULT_START_RATING_SEARCH_RANGE = [
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='int',
        low=50,
        high=200,
    ),
    ParameterSearchRange(
        name='team_rating_subtract',
        type='int',
        low=20,
        high=300
    ),
    ParameterSearchRange(
        name='team_weight',
        type='uniform',
        low=0,
        high=0.7
    )
]


class RatingGeneratorTuner(ABC):

    @abstractmethod
    def tune(self, df: pd.DataFrame, rating_idx: int, scorer: BaseScorer,
             match_predictor_factory: MatchPredictorFactory,
             matches: list[Match]) -> RatingGenerator:
        pass


class OpponentAdjustedRatingGeneratorTuner(RatingGeneratorTuner):

    def __init__(self,
                 team_rating_search_ranges: Optional[list[ParameterSearchRange]] = None,
                 team_rating_n_trials: int = 30,
                 start_rating_search_ranges: Optional[list[ParameterSearchRange]] = None,
                 start_rating_n_trials: int = 8,
                 ):
        self.team_rating_search_ranges = team_rating_search_ranges or DEFAULT_TEAM_SEARCH_RANGES
        self.start_rating_search_ranges = start_rating_search_ranges or DEFAULT_START_RATING_SEARCH_RANGE
        self.team_rating_n_trials = team_rating_n_trials
        self.start_rating_n_trials = start_rating_n_trials

    def tune(self,
             df: pd.DataFrame,
             rating_idx: int,
             scorer: BaseScorer,
             match_predictor_factory: MatchPredictorFactory,
             matches: list[Match]) -> OpponentAdjustedRatingGenerator:

        if match_predictor_factory.rating_generators:
            best_rating_generator = copy.deepcopy(match_predictor_factory.rating_generators[rating_idx])
        else:
            potential_rating_features = [v for k, v in RatingColumnNames.__dict__.items() if isinstance(v, str)]

            best_rating_generator = OpponentAdjustedRatingGenerator(
                features_out=[f for f in match_predictor_factory.predictor.features if f in potential_rating_features],
                column_names=match_predictor_factory.rating_generators[rating_idx].column_names
            )

        if self.team_rating_n_trials > 0:
            logging.info("Tuning Team Rating")
            best_team_rating_generator = self._tune_team_rating(df=df,
                                                                rating_generator=best_rating_generator,
                                                                rating_index=rating_idx,
                                                                scorer=scorer,
                                                                matches=matches,
                                                                match_predictor_factory=match_predictor_factory)

            best_rating_generator.team_rating_generator = best_team_rating_generator

        if self.start_rating_n_trials > 0:
            logging.info("Tuning Start Rating")

            best_start_rating = self._tune_start_rating(df=df,
                                                        matches=matches,
                                                        rating_generator=best_rating_generator,
                                                        rating_index=rating_idx, scorer=scorer,
                                                        match_predictor_factory=match_predictor_factory
                                                        )
            best_rating_generator.team_rating_generator.start_rating_generator = best_start_rating

        return OpponentAdjustedRatingGenerator(team_rating_generator=best_rating_generator.team_rating_generator,
                                               column_names=best_rating_generator.column_names)

    def _tune_team_rating(self,
                          df: pd.DataFrame,
                          rating_generator: OpponentAdjustedRatingGenerator,
                          rating_index: int,
                          matches: list[Match],
                          scorer: BaseScorer,
                          match_predictor_factory: MatchPredictorFactory,
                          ) -> TeamRatingGenerator:

        def objective(trial: BaseTrial, df: pd.DataFrame) -> float:

            team_rating_generator_params = list(
                inspect.signature(rating_generator.team_rating_generator.__class__.__init__).parameters.keys())[1:]

            params = {attr: getattr(rating_generator.team_rating_generator, attr) for attr in
                      team_rating_generator_params if attr not in ('performance_predictor', 'start_rating_generator')}
            params = add_params_from_search_range(params=params,
                                                  trial=trial,
                                                  parameter_search_range=self.team_rating_search_ranges)

            performance_predictor = rating_generator.team_rating_generator.performance_predictor
            performance_predictor_params = list(
                inspect.signature(performance_predictor.__class__.__init__).parameters.keys())[1:]

            for param in params.copy():
                if param in performance_predictor_params:
                    performance_predictor.__setattr__(param, params[param])
                    params.pop(param)

            team_rating_generator = TeamRatingGenerator(**params,
                                                        performance_predictor=performance_predictor)

            rating_g = copy.deepcopy(rating_generator)
            rating_g.team_rating_generator = team_rating_generator
            rating_generators = copy.deepcopy(match_predictor_factory.rating_generators)

            rating_generators[rating_index] = rating_g
            match_predictor = match_predictor_factory.create(
                rating_generators=rating_generators,
            )

            df_with_prediction = match_predictor.generate_historical(df=df, matches=matches, store_ratings=False)
            test_df = df_with_prediction[df_with_prediction[match_predictor.date_column_name] > match_predictor.train_split_date]
            return scorer.score(test_df, classes_=match_predictor.predictor.classes_)

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(direction=direction, study_name=study_name, sampler=sampler)
        callbacks = []
        study.optimize(lambda trial: objective(trial, df), n_trials=self.team_rating_n_trials, callbacks=callbacks)

        best_params = study.best_params
        team_rating_generator_params = list(
            inspect.signature(rating_generator.team_rating_generator.__class__.__init__).parameters.keys())[1:]
        other_params = {attr: getattr(rating_generator.team_rating_generator, attr) for attr in
                        team_rating_generator_params
                        if attr not in ('performance_predictor', 'start_rating_generator')}
        for param in other_params:
            if param not in best_params:
                best_params[param] = other_params[param]

        performance_predictor = rating_generator.team_rating_generator.performance_predictor
        performance_predictor_params = list(
            inspect.signature(performance_predictor.__class__.__init__).parameters.keys())[1:]

        for param in best_params.copy():
            if param in performance_predictor_params:
                performance_predictor.__setattr__(param, best_params[param])
                best_params.pop(param)

        return TeamRatingGenerator(**best_params,
                                   performance_predictor=performance_predictor,
                                   start_rating_generator=rating_generator.team_rating_generator.start_rating_generator
                                   )

    def _tune_start_rating(self,
                           df: pd.DataFrame,
                           rating_generator: OpponentAdjustedRatingGenerator,
                           rating_index: int,
                           matches: list[Match],
                           scorer: BaseScorer,
                           match_predictor_factory: MatchPredictorFactory):
        def objective(trial: BaseTrial, df: pd.DataFrame) -> float:
            start_rating_generator_params = list(
                inspect.signature(
                    rating_generator.team_rating_generator.start_rating_generator.__class__.__init__).parameters.keys())[
                                            1:]

            params = {attr: getattr(rating_generator.team_rating_generator.start_rating_generator, attr) for attr in
                      start_rating_generator_params}

            params = add_params_from_search_range(params=params,
                                                  trial=trial,
                                                  parameter_search_range=self.start_rating_search_ranges)

            start_rating_generator = StartRatingGenerator(**params)
            rating_g = copy.deepcopy(rating_generator)
            rating_g.team_rating_generator.start_rating_generator = start_rating_generator
            if match_predictor_factory.rating_generators:
                rating_generators = copy.deepcopy(match_predictor_factory.rating_generators)
                rating_generators[rating_index] = rating_g
            else:
                rating_generators = [rating_g]
            match_predictor = match_predictor_factory.create(
                rating_generators=rating_generators,

            )

            df_with_prediction = match_predictor.generate_historical(df=df, matches=matches, store_ratings=False)
            test_df = df_with_prediction[
                df_with_prediction[match_predictor.date_column_name] > match_predictor.train_split_date]
            return scorer.score(test_df, classes_=match_predictor.predictor.classes_)

        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(direction=direction, study_name=study_name, sampler=sampler)
        callbacks = []
        study.optimize(lambda trial: objective(trial, df), n_trials=self.start_rating_n_trials, callbacks=callbacks)
        start_rating_generator_params = list(
            inspect.signature(
                rating_generator.team_rating_generator.start_rating_generator.__class__.__init__).parameters.keys())[1:]

        other_params = {attr: getattr(rating_generator.team_rating_generator.start_rating_generator, attr) for attr in
                        start_rating_generator_params}

        best_params = study.best_params
        for param in other_params:
            if param not in best_params:
                best_params[param] = other_params[param]

        return StartRatingGenerator(**best_params)



from sklearn.preprocessing import StandardScaler

from player_performance_ratings import ColumnNames
from player_performance_ratings.ratings.performances_generator import ColumnWeight, PerformancesGenerator

from player_performance_ratings.transformation import SkLearnTransformerWrapper, MinMaxTransformer

from player_performance_ratings.transformation.pre_transformers import NetOverPredictedTransformer, \
    SymmetricDistributionTransformer


def auto_create_performance_generator(column_weights: list[list[ColumnWeight]], column_names: list[ColumnNames],
                                      ) -> PerformancesGenerator:
    """
    Creates a list of transformers that ensure the performance column is generated in a way that makes sense for the rating model.
    Ensures columns aren't too skewed, scales them to similar ranges, ensure values are between 0 and 1,
    and then weights them according to the column_weights.
    """
    pre_transformations = []
    if not isinstance(column_weights[0], list):
        column_weights = [column_weights]

    all_feature_names = []
    for col_weights in column_weights:
        for col_weight in col_weights:
            if col_weight.name not in all_feature_names:
                all_feature_names.append(col_weight.name)



    contains_position = True if any([c.position is not None for c in column_names]) else False
    contains_not_position = True if any([c.position is None for c in column_names]) else False

    not_position_features = []
    position_features = []
    if contains_position:
        for idx, col_weights in enumerate(column_weights):
            feature_names = []
            if column_names[idx].position is None:
                granularity = []
            else:
                granularity = [column_names[idx].position]

            for column_weight in col_weights:
                feature = column_weight.name
                feature_names.append(feature)

            if column_names[idx].position is not None:
                feats = []
                for col in column_weights[idx]:
                    if col.name not in feats:
                        feats.append(col.name)
                position_predicted_transformer = NetOverPredictedTransformer(features=feats,
                                                                             granularity=[column_names[idx].position],
                                                                             prefix="net_position_predicted__")
                pre_transformations.append(position_predicted_transformer)

                distribution_transformer = SymmetricDistributionTransformer(features=position_predicted_transformer.features_out,
                                                                            granularity=granularity,
                                                                            prefix="symmetric_position__")
                pre_transformations.append(distribution_transformer)
                position_features  += [f for f in distribution_transformer.features_out if f not in position_features]
                for idx2, col_weight in enumerate(column_weights[idx]):
                    column_weights[idx][idx2].name = distribution_transformer.prefix + position_predicted_transformer.prefix  +col_weight.name

            else:
                not_position_features += [c.name for c in column_weights[idx]]

        all_feature_names = not_position_features + position_features
    else:
        not_position_features = all_feature_names

    if contains_not_position:
        distribution_transformer = SymmetricDistributionTransformer(features=not_position_features)
        pre_transformations.append(distribution_transformer)

    pre_transformations.append(
        SkLearnTransformerWrapper(transformer=StandardScaler(), features=all_feature_names))

    pre_transformations.append(MinMaxTransformer(features=all_feature_names))

    return PerformancesGenerator(column_names=column_names, pre_transformations=pre_transformations,
                                 column_weights=column_weights)



import pickle

import pandas as pd
from lightgbm import LGBMRegressor
from player_performance_ratings.scorer.score import SklearnScorer
from sklearn.metrics import mean_absolute_error
from player_performance_ratings.tuner import MatchPredictorTuner

from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory

from player_performance_ratings.transformation import LagTransformer

from player_performance_ratings import ColumnNames, PredictColumnNames

from player_performance_ratings.transformation.post_transformers import NormalizerTransformer
from player_performance_ratings.tuner.predictor_tuner import PredictorTuner
from player_performance_ratings.tuner.utils import get_default_lgbm_regressor_search_range_by_learning_rate

df = pd.read_pickle(r"data/game_player_full.pickle")
df = df.drop_duplicates(subset=['game_id', 'player_id'])
df = (
    df.assign(team_count=df.groupby("game_id")["team_id"].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
    .drop(columns=['team_count'])
)

df.loc[df['start_position'] != '', 'starting'] = 1
df.loc[df['start_position'] == '', 'starting'] = 0

df['is_playoff'] = 0
df.loc[
    (df['start_date'].between('2018-04-13', '2018-06-20')) |
    (df['start_date'].between('2019-04-13', '2019-06-27')) |
    (df['start_date'].between('2020-08-13', '2020-09-20')) |
    (df['start_date'].between('2021-05-20', '2021-07-22')) |
    (df['start_date'].between('2022-04-13', '2022-06-27')) |
    (df['start_date'].between('2023-04-13', '2023-06-27')), 'is_playoff'
] = 1

df[PredictColumnNames.TARGET] = df['minutes']

df = df.sort_values(["start_date", "game_id", "team_id", "player_id"])
column_names = ColumnNames(
    team_id='team_id',
    match_id='game_id',
    start_date="start_date",
    player_id="player_id",
    performance=None,
)

lag_transformer = LagTransformer(
    features=["minutes"],
    lag_length=20,
    granularity=[column_names.player_id, "starting"],
    column_names=column_names,
)

train_split_date = df.iloc[int(len(df) / 1.3)][column_names.start_date]

post_rating_transformers = [
    LagTransformer(
        features=["minutes"],
        lag_length=25,
        granularity=[column_names.player_id, "starting"],
        column_names=column_names,
        days_between_lags=[1, 2, 3, 10, 25]
    ),
    LagTransformer(
        features=["minutes"],
        lag_length=10,
        granularity=[column_names.player_id, "starting", "is_playoff"],
        column_names=column_names,
        days_between_lags=[1, 2, 4],
        prefix="is_playoff_lag_"
    ),
    LagTransformer(
        features=["minutes"],
        lag_length=1,
        granularity=[column_names.player_id, "starting", 'location'],
        column_names=column_names,
        prefix="starting_location_lag_"
    ),
    LagTransformer(
        features=[],
        lag_length=1,
        granularity=[column_names.player_id],
        column_names=column_names,
        prefix="future_lag",
        future_lag=True,
        days_between_lags=[2]
    ),
]

post_prediction_transformers = [
    NormalizerTransformer(features=[f"{PredictColumnNames.TARGET}_prediction"],
                          granularity=[column_names.team_id, column_names.match_id],
                          create_target_as_mean=True)]

match_predictor_factory = MatchPredictorFactory(
    post_rating_transformers=post_rating_transformers,
    post_prediction_transformers=post_prediction_transformers,
    estimator=LGBMRegressor(reg_alpha=1, learning_rate=0.02, verbose=-100),
    date_column_name=column_names.start_date,
    other_categorical_features=["starting", "is_playoff"],
)

predictor_tuner = PredictorTuner(
    search_ranges=get_default_lgbm_regressor_search_range_by_learning_rate(
        learning_rate=match_predictor_factory.estimator.learning_rate),
    n_trials=65,
    date_column_name=column_names.start_date,
)

match_tuner = MatchPredictorTuner(
    scorer=SklearnScorer(target=PredictColumnNames.TARGET,
                         scorer_function=mean_absolute_error,
                         pred_column=match_predictor_factory.predictor.pred_column),
    match_predictor_factory=match_predictor_factory,
    predictor_tuner=predictor_tuner
)

best_model = match_tuner.tune(df=df)

df_with_minutes_prediction = best_model.generate_historical(df=df)

pickle.dump(best_model, open("models/nba_minute_prediction", 'wb'))
df_with_minutes_prediction.to_pickle("data/game_player_full_with_minutes_prediction.pickle")



import pandas as pd
from lightgbm import LGBMRegressor
from sklearn.metrics import mean_absolute_error

from player_performance_ratings import ColumnNames

from player_performance_ratings.predictor import MatchPredictor
from player_performance_ratings.ratings import BayesianTimeWeightedRating, ColumnWeight

df = pd.read_parquet("data/subsample_lol_data")
df = df.sort_values(by=['date', 'gameid', 'teamname', "playername"])
df['__target'] = df['kills']
df["__target"] = df["__target"].clip(0, 8)

df['kills_per_minute'] = df['kills'] / df['gamelength'] * 60

df = (
    df.loc[lambda x: x.position != 'team']
    .assign(team_count=df.groupby('gameid')['teamname'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)
column_names_kpm = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='kills_per_minute',
    league='league',
    position='position',
)

column_names_kills = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='kills',
    league='league',
    position='position'
)
time_weighed_rating_kills_per_minute = BayesianTimeWeightedRating(column_names=column_names_kpm)
time_weighed_rating_kills = BayesianTimeWeightedRating(column_names=column_names_kills)



match_predictor = MatchPredictor(
    rating_generators=[time_weighed_rating_kills_per_minute, time_weighed_rating_kills],
    use_auto_create_performance_calculator=True,
    estimator=LGBMRegressor(),
    column_weights=[[ColumnWeight(name='kills_per_minute', weight=1)], [ColumnWeight(name='kills', weight=1)]],
    other_categorical_features=["position"]

)

df_predictions = match_predictor.generate_historical(df=df)

print(mean_absolute_error(df_predictions['__target'], df_predictions[match_predictor.predictor.pred_column]))



import pickle

import pandas as pd
from lightgbm import LGBMClassifier
from player_performance_ratings.transformation.pre_transformers import SymmetricDistributionTransformer, \
    NetOverPredictedTransformer

from player_performance_ratings.predictor.estimators import Predictor
from player_performance_ratings.scorer import LogLossScorer
from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory
from player_performance_ratings.tuner.predictor_tuner import PredictorTuner

from player_performance_ratings.tuner.rating_generator_tuner import OpponentAdjustedRatingGeneratorTuner

from player_performance_ratings.ratings import OpponentAdjustedRatingGenerator, RatingColumnNames

from player_performance_ratings.transformation import SkLearnTransformerWrapper, MinMaxTransformer
from sklearn.preprocessing import StandardScaler

from player_performance_ratings.data_structures import ColumnNames

from player_performance_ratings.tuner import MatchPredictorTuner, PerformancesGeneratorTuner
from player_performance_ratings.tuner.utils import ParameterSearchRange



column_names = ColumnNames(
    team_id='teamname',
    match_id='gameid',
    start_date="date",
    player_id="playername",
    performance='performance',
    league='league',
    position='champion_position',
)
df = pd.read_parquet("data/subsample_lol_data")
df = df.sort_values(by=['date', 'gameid', 'teamname', "playername"])
df['champion_position'] = df['champion'] + df['position']
df['__target'] = df['result']

df = (
    df.loc[lambda x: x.position != 'team']
    .assign(team_count=df.groupby('gameid')['teamname'].transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)

rating_generator = OpponentAdjustedRatingGenerator(column_names=column_names)

team_rating_search_ranges = [
    ParameterSearchRange(
        name='confidence_weight',
        type='uniform',
        low=0.7,
        high=0.95
    ),
    ParameterSearchRange(
        name='confidence_days_ago_multiplier',
        type='uniform',
        low=0.02,
        high=.12,
    ),
    ParameterSearchRange(
        name='confidence_max_days',
        type='uniform',
        low=40,
        high=150,
    ),
    ParameterSearchRange(
        name='confidence_max_sum',
        type='uniform',
        low=60,
        high=300,
    ),
    ParameterSearchRange(
        name='confidence_value_denom',
        type='uniform',
        low=50,
        high=350
    ),
    ParameterSearchRange(
        name='rating_change_multiplier',
        type='uniform',
        low=30,
        high=100
    ),
    ParameterSearchRange(
        name='min_rating_change_multiplier_ratio',
        type='uniform',
        low=0.02,
        high=0.2,
    )
]

weighter_search_range = {
    column_names.performance:
        [
            ParameterSearchRange(
                name='damagetochampions',
                type='uniform',
                low=0,
                high=0.45
            ),
            ParameterSearchRange(
                name='deaths',
                type='uniform',
                low=0,
                high=.3,
            ),
            ParameterSearchRange(
                name='kills',
                type='uniform',
                low=0,
                high=0.3
            ),
            ParameterSearchRange(
                name='result',
                type='uniform',
                low=0.25,
                high=0.85
            ),
        ]
}

features = ["damagetochampions", "result",
            "kills", "deaths"]

start_rating_search_range = [
    ParameterSearchRange(
        name='team_weight',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.12,
        high=.4,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='uniform',
        low=20,
        high=100,
    )
]

position_based_features = ["kills", "deaths", "damagetochampions"]


pre_transformations = [
    NetOverPredictedTransformer(features=position_based_features,
                                granularity=[column_names.position],
                                prefix=""),
    SymmetricDistributionTransformer(features=position_based_features),
    SkLearnTransformerWrapper(transformer=StandardScaler(), features=position_based_features),
    MinMaxTransformer(features=position_based_features)
]

performance_generator_tuner = PerformancesGeneratorTuner(performances_weight_search_ranges=weighter_search_range,
                                                         pre_transformations=pre_transformations,
                                                         lower_is_better_features=["deaths"],
                                                         n_trials=20
                                                         )

rating_generator_tuner = OpponentAdjustedRatingGeneratorTuner(
    team_rating_search_ranges=team_rating_search_ranges,
    start_rating_search_ranges=start_rating_search_range,
)

predictor_tuner = PredictorTuner(
    search_ranges=[
        ParameterSearchRange(
            name='n_estimators',
            type='int',
            low=50,
            high=250,
        ),
        ParameterSearchRange(
            name='num_leaves',
            type='int',
            low=10,
            high=100,
        ),
        ParameterSearchRange(
            name='max_depth',
            type='int',
            low=2,
            high=7,
        ),
        ParameterSearchRange(
            name='min_child_samples',
            type='int',
            low=10,
            high=100,
        ),
        ParameterSearchRange(
            name='reg_alpha',
            type='uniform',
            low=0,
            high=5,
        ),
    ],
    date_column_name="date",
)

match_predictor_factory = MatchPredictorFactory(
    rating_generators=rating_generator,
    predictor=Predictor(
        estimator=LGBMClassifier(verbose=-100),
        features=[RatingColumnNames.RATING_DIFFERENCE_PROJECTED],
    ),
    date_column_name="date",
)

scorer = LogLossScorer(pred_column=match_predictor_factory.predictor.pred_column)

tuner = MatchPredictorTuner(
    performances_generator_tuner=performance_generator_tuner,
    rating_generator_tuners=rating_generator_tuner,
    predictor_tuner=predictor_tuner,
    fit_best=True,
    match_predictor_factory=match_predictor_factory,
    scorer=scorer
)
best_match_predictor = tuner.tune(df=df)
pickle.dump(best_match_predictor, open("models/lol_match_predictor", 'wb'))



import numpy as np

from player_performance_ratings.transformation import SkLearnTransformerWrapper, MinMaxTransformer
from player_performance_ratings.tuner.utils import ParameterSearchRange
from sklearn.preprocessing import StandardScaler

from player_performance_ratings.examples.utils import load_nba_game_matchup_data

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.data_structures import ColumnNames


from player_performance_ratings.predictor.estimators.classifier import SkLearnGameTeamPredictor

from player_performance_ratings.ratings.enums import RatingColumnNames
from player_performance_ratings.ratings import TeamRatingGenerator
from player_performance_ratings.ratings.opponent_adjusted_rating.performance_predictor import RatingDifferencePerformancePredictor
from player_performance_ratings.ratings.opponent_adjusted_rating.start_rating_generator import StartRatingGenerator
from player_performance_ratings.ratings.rating_generator import OpponentAdjustedRatingGenerator
from player_performance_ratings.tuner import MatchPredictorTuner
from player_performance_ratings.scorer import LogLossScorer


from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory
from player_performance_ratings.tuner.rating_generator_tuner import OpponentAdjustedRatingGeneratorTuner

column_names = ColumnNames(
    team_id='lineup_id',
    match_id='matchup_game_id',
    start_date="start_date",
    player_id="player_id",
    performance="plus_minus_per_minute",
    participation_weight="participation_weight",
    rating_update_id="game_id"

)
df = load_nba_game_matchup_data()
df.loc[df['points'] > df['points_opponent'], column_names.performance] = 1
df.loc[df['points'] < df['points_opponent'], column_names.performance] = 0
df.loc[df['points'] == df['points_opponent'], column_names.performance] = 0.5
df[PredictColumnNames.TARGET] = df['won']

min_lineup = np.minimum(df['lineup_id'], df['lineup_id_opponent'])
max_lineup = np.maximum(df['lineup_id'], df['lineup_id_opponent'])

# Combine the min and max values into a tuple
df['sorted_lineup'] = list(zip(min_lineup, max_lineup))
df = df[df['game_minutes'] > 46]

df[column_names.match_id] = df['game_id'].astype(str) + '_' + df['sorted_lineup'].astype(str)
df = df.drop(columns=['sorted_lineup'])

df = (
    df.assign(team_count=df.granularity(column_names.match_id)[column_names.team_id].fit_transform('nunique'))
    .loc[lambda x: x.team_count == 2]
)
df = df.sort_values(by=[column_names.start_date, column_names.match_id, column_names.team_id, column_names.player_id])
df['plus_minus'] = df['points'] - df['points_opponent']
df['plus_minus_per_minute'] = df['plus_minus'] / df['minutes_lineup_matchup']
df.loc[df['minutes_lineup_matchup'] == 0, 'plus_minus_per_minute'] = 0
print(len(df['game_id'].unique()))
mean_participation_weight = df['participation_weight'].mean()
features = ['plus_minus_per_minute']

standard_scaler = SkLearnTransformerWrapper(transformer=StandardScaler(), features=features)
pre_transformer_search_ranges = [
    (standard_scaler, []),
    (MinMaxTransformer(features=features), [])
]

team_rating_search_ranges = [
    ParameterSearchRange(
        name='certain_weight',
        type='uniform',
        low=0.7,
        high=0.95
    ),
    ParameterSearchRange(
        name='certain_days_ago_multiplier',
        type='uniform',
        low=0.02,
        high=.12,
    ),
    ParameterSearchRange(
        name='max_days_ago',
        type='uniform',
        low=40,
        high=200,
    ),
    ParameterSearchRange(
        name='max_certain_sum',
        type='uniform',
        low=20,
        high=70,
    ),
    ParameterSearchRange(
        name='certain_value_denom',
        type='uniform',
        low=10,
        high=50
    ),
    ParameterSearchRange(
        name='reference_certain_sum_value',
        type='uniform',
        low=0.4,
        high=5
    ),
    ParameterSearchRange(
        name='rating_change_multiplier',
        type='uniform',
        low=35,
        high=140
    ),
    ParameterSearchRange(
        name='participation_weight_coef',
        type='uniform',
        low=0,
        high=10
    ),
    ParameterSearchRange(
        name='team_rating_diff_coef',
        type='uniform',
        low=0,
        high=0.002
    ),
]

performance_predictor = RatingDifferencePerformancePredictor(
    team_rating_diff_coef=0,
    rating_diff_coef=0.005757,
    participation_weight_coef=1,
    mean_participation_weight=mean_participation_weight

)

rating_generator = OpponentAdjustedRatingGenerator(
    team_rating_generator=TeamRatingGenerator(
        performance_predictor=performance_predictor,
        start_rating_generator=StartRatingGenerator(
            team_weight=0,
        )
    )
)

pre_transformers = [
    standard_scaler,
    MinMaxTransformer(features=features, quantile=0.99),
]

start_rating_search_range = [
    ParameterSearchRange(
        name='league_quantile',
        type='uniform',
        low=0.04,
        high=.4,
    ),
    ParameterSearchRange(
        name='min_count_for_percentiles',
        type='int',
        low=50,
        high=400,
    ),
]

predictor = SkLearnGameTeamPredictor(features=[RatingColumnNames.RATING_DIFFERENCE],
                                     weight_column='participation_weight',
                                     team_id_column='team_id', game_id_colum=column_names.rating_update_id,
                                     target='won')

match_predictor_factory = MatchPredictorFactory(
    column_names=column_names,
    rating_generators=rating_generator,
    pre_transformers=pre_transformers,
    predictor=predictor,
    train_split_date="2022-05-01"
)



rating_generator_tuner = OpponentAdjustedRatingGeneratorTuner(
    start_rating_search_ranges=start_rating_search_range,
    start_rating_n_trials=3,
    team_rating_search_ranges=team_rating_search_ranges,
    team_rating_n_trials=3,

)

tuner = MatchPredictorTuner(
    match_predictor_factory=match_predictor_factory,
    fit_best=True,
    scorer=LogLossScorer(pred_column=predictor.pred_column),
    rating_generator_tuners=rating_generator_tuner,

)
best_match_predictor = tuner.tune(df=df)
tuner.tune(df=df)



import copy
import inspect
from typing import Optional

import optuna
import pandas as pd
import pendulum
from optuna.samplers import TPESampler
from optuna.trial import BaseTrial

from player_performance_ratings.data_structures import Match
from player_performance_ratings.tuner.match_predictor_factory import MatchPredictorFactory

from player_performance_ratings.predictor.estimators.base_estimator import BaseMLWrapper
from player_performance_ratings.scorer import BaseScorer

from player_performance_ratings.tuner.utils import ParameterSearchRange, add_params_from_search_range


class PredictorTuner():

    def __init__(self,
                 search_ranges: list[ParameterSearchRange],
                 date_column_name: str,
                 train_split_date: Optional[pendulum.datetime] = None,
                 default_params: Optional[dict] = None,
                 estimator_subclass_level: int = 0,
                 n_trials: int = 30
                 ):
        self.search_ranges = search_ranges
        self.date_column_name = date_column_name
        self.train_split_date = train_split_date
        self.default_params = default_params or {}
        self.estimator_subclass_level = estimator_subclass_level
        self.n_trials = n_trials

    def tune(self, df: pd.DataFrame,  scorer: BaseScorer,
             match_predictor_factory: MatchPredictorFactory) -> BaseMLWrapper:

        if not self.train_split_date:
            self.train_split_date = df.iloc[int(len(df) / 1.3)][self.date_column_name]

        def objective(trial: BaseTrial, df: pd.DataFrame) -> float:

            predictor = copy.deepcopy(match_predictor_factory.predictor)

            if self.estimator_subclass_level == 0:
                param_names = list(
                    inspect.signature(predictor.estimator.__class__.__init__).parameters.keys())[1:]
                params = {attr: getattr(predictor.estimator, attr) for attr in param_names if attr != 'kwargs'}
                if '_other_params' in predictor.estimator.__dict__:
                    params.update(predictor.estimator._other_params)
            elif self.estimator_subclass_level == 1:
                param_names = list(
                    inspect.signature(predictor.estimator.estimator.__class__.__init__).parameters.keys())[1:]
                params = {attr: getattr(predictor.estimator.estimator, attr) for attr in param_names if attr != 'kwargs'}
                if '_other_params' in predictor.estimator.estimator.__dict__:
                    params.update(predictor.estimator.estimator._other_params)
            elif self.estimator_subclass_level == 2:
                param_names = list(
                    inspect.signature(predictor.estimator.estimator.estimator.__class__.__init__).parameters.keys())[1:]
                params = {attr: getattr(predictor.estimator.estimator.estimator, attr) for attr in param_names if attr != 'kwargs'}
                if '_other_params' in predictor.estimator.estimator.estimator.__dict__:
                    params.update(predictor.estimator.estimator.estimator._other_params)

            else:
                raise ValueError(f"estimator_subclass_level can't be higher than 2, got {self.estimator_subclass_level}")


            params = add_params_from_search_range(params=params,
                                                  trial=trial,
                                                  parameter_search_range=self.search_ranges)
            for param, value in self.default_params.items():
                params[param] = value

            predictor = copy.deepcopy(match_predictor_factory.predictor)
            for param in params:
                if self.estimator_subclass_level == 1:
                    setattr(predictor.estimator.estimator, param, params[param])
                elif self.estimator_subclass_level == 2:
                    setattr(predictor.estimator.estimator.estimator, param, params[param])
                elif self.estimator_subclass_level >2:
                    raise ValueError(f"estimator_subclass_level can't be higher than 2, got {self.estimator_subclass_level}")
                else:
                    setattr(predictor.estimator, param, params[param])
            train_df = df[df[self.date_column_name] <= self.train_split_date]
            predict_df = df[df[self.date_column_name] > self.train_split_date]

            predictor.train(train_df)
            df_with_prediction = predictor.add_prediction(predict_df)
            return scorer.score(df_with_prediction, classes_=match_predictor_factory.predictor.classes_)



        direction = "minimize"
        study_name = "optuna_study"
        optuna_seed = 12
        sampler = TPESampler(seed=optuna_seed)
        study = optuna.create_study(direction=direction, study_name=study_name, sampler=sampler)
        callbacks = []
        study.optimize(lambda trial: objective(trial, df), n_trials=self.n_trials, callbacks=callbacks)
        best_estimator_params = study.best_params
        other_predictor_params = list(
            inspect.signature(match_predictor_factory.predictor.__class__.__init__).parameters.keys())[1:]

        if self.estimator_subclass_level > 0:
            if self.estimator_subclass_level == 1:
                best_estimator_params.update(match_predictor_factory.predictor.estimator.estimator._other_params)
            elif self.estimator_subclass_level == 2:
                best_estimator_params.update(match_predictor_factory.predictor.estimator.estimator.estimator._other_params)

        else:
            if '_other_params' in match_predictor_factory.predictor.estimator.__dict__:
                best_estimator_params.update(match_predictor_factory.predictor.estimator._other_params)


        other_predictor_params = {attr: getattr(match_predictor_factory.predictor, attr) for attr in
                                  other_predictor_params if attr not in ('estimator')}

        predictor_class = match_predictor_factory.predictor.__class__
        if self.estimator_subclass_level == 1:

            potential_parent_names = list(
                inspect.signature(
                    match_predictor_factory.predictor.estimator.__class__.__init__).parameters.keys())[1:]
            other_parent_params = {attr: getattr(match_predictor_factory.predictor.estimator, attr) for attr
                                   in
                                   potential_parent_names if attr != 'estimator'}

            estimator_class = match_predictor_factory.predictor.estimator.estimator.__class__
            parent_estimator_class = match_predictor_factory.predictor.estimator.__class__
            parent_estimator = parent_estimator_class(estimator=estimator_class(**best_estimator_params), **other_parent_params)
            return predictor_class(estimator=parent_estimator, **other_predictor_params)
        elif self.estimator_subclass_level == 2:
            potential_parent_names = list(
                inspect.signature(
                    match_predictor_factory.predictor.estimator.estimator.__class__.__init__).parameters.keys())[1:]
            other_parent_params = {attr: getattr(match_predictor_factory.predictor.estimator.estimator, attr) for attr
                                   in
                                   potential_parent_names if attr != 'estimator'}
            estimator_class = match_predictor_factory.predictor.estimator.estimator.estimator.__class__
            parent_estimator_class = match_predictor_factory.predictor.estimator.estimator.__class__
            parent_estimator = parent_estimator_class(estimator=estimator_class(**best_estimator_params),
                                                      **other_parent_params)

            parent_parent_estimator_class = match_predictor_factory.predictor.estimator.__class__
            parent_parent_estimator = parent_parent_estimator_class(estimator=parent_estimator)
            return predictor_class(estimator=parent_parent_estimator, **other_predictor_params)

        elif self.estimator_subclass_level == 0:

            estimator_class = match_predictor_factory.predictor.estimator.__class__
            return predictor_class(estimator=estimator_class(**best_estimator_params), **other_predictor_params)

        else:
            raise ValueError(f"estimator_subclass_level can't be higher than 2, got {self.estimator_subclass_level}")



import logging
from typing import List, Optional, Union

import pandas as pd
import pendulum
from sklearn.preprocessing import OneHotEncoder

from player_performance_ratings.ratings import PerformancesGenerator, ColumnWeight

from player_performance_ratings.consts import PredictColumnNames
from player_performance_ratings.predictor.estimators.base_estimator import BaseMLWrapper

from player_performance_ratings.predictor.estimators import Predictor, GameTeamPredictor
from player_performance_ratings.data_structures import Match
from player_performance_ratings.ratings.league_identifier import LeagueIdentifier
from player_performance_ratings.ratings.match_generator import convert_df_to_matches
from player_performance_ratings.ratings.rating_generator import RatingGenerator

from player_performance_ratings.transformation.base_transformer import BaseTransformer, BasePostTransformer
from player_performance_ratings.transformation.factory import auto_create_performance_generator

from player_performance_ratings.transformation.pre_transformers import ConvertDataFrameToCategoricalTransformer, \
    SkLearnTransformerWrapper

def create_predictor(
        rating_generators: Optional[Union[RatingGenerator, list[RatingGenerator]]],
        other_features: Optional[list[str]],
        other_categorical_features: Optional[list[str]],
        post_rating_transformers: Optional[List[BasePostTransformer]],
        estimator: Optional,
        group_predictor_by_game_team: bool,
        match_id_column_name: Optional[str],
        team_id_column_name: Optional[str] ,

) -> BaseMLWrapper:
    features = list(set(other_features + other_categorical_features)) or []

    for c in post_rating_transformers:
        features += c.features_out
    for rating_idx, c in enumerate(rating_generators):
        for rating_feature in c.features_out:
            if len(rating_generators) > 1:
                rating_feature_str = rating_feature + str(rating_idx)
            else:
                rating_feature_str = rating_feature
            features.append(rating_feature_str)

    logging.warning(f"predictor is not set. Will use {features} as features")
    if other_categorical_features:
        if estimator.__class__.__name__ in ('LGBMClassifier', 'LGBMRegressor', "XGBClassifier", "XGBRegressor"):
            categorical_transformers = [
                ConvertDataFrameToCategoricalTransformer(features=other_categorical_features)
            ]
        else:
            categorical_transformers = [
                SkLearnTransformerWrapper(
                    transformer=OneHotEncoder(handle_unknown='ignore'), features=other_categorical_features)
            ]
    else:
        categorical_transformers = []

    if group_predictor_by_game_team:
        match_id = rating_generators[
            0].column_names.match_id if rating_generators else match_id_column_name
        team_id = rating_generators[0].column_names.team_id if rating_generators else team_id_column_name
        if match_id is None or team_id is None:
            raise ValueError(
                "match_id and team_id must be set if group_predictor_by_game_team is used to create predictor")

        return  GameTeamPredictor(
            estimator=estimator,
            features=features,
            target=PredictColumnNames.TARGET,
            game_id_colum=match_id,
            team_id_column=team_id,
            categorical_transformers=categorical_transformers
        )


    else:
        return  Predictor(
            estimator=estimator,
            features=features,
            target=PredictColumnNames.TARGET,
            categorical_transformers=categorical_transformers
        )


class MatchPredictor():

    def __init__(self,
                 rating_generators: Optional[Union[RatingGenerator, list[RatingGenerator]]] = None,
                 performances_generator: Optional[PerformancesGenerator] = None,
                 post_rating_transformers: Optional[List[BasePostTransformer]] = None,
                 post_prediction_transformers: Optional[List[BasePostTransformer]] = None,
                 predictor: [Optional[BaseMLWrapper]] = None,
                 estimator: Optional = None,
                 estimator_or_transformers: Optional[Union[BaseMLWrapper, List[BaseTransformer]]] = None,
                 other_features: Optional[list[str]] = None,
                 other_categorical_features: Optional[list[str]] = None,
                 group_predictor_by_game_team: bool = False,
                 train_split_date: Optional[pendulum.datetime] = None,
                 date_column_name: Optional[str] = None,
                 match_id_column_name: Optional[str] = None,
                 team_id_column_name: Optional[str] = None,
                 use_auto_create_performance_calculator: bool = False,
                 column_weights: Optional[Union[list[list[ColumnWeight]], list[ColumnWeight]]] = None,
                 keep_features: bool = False,
                 ):

        """

        :param column_names:
        :param rating_generators:
        A single or a list of RatingGenerators.

        :param performances_generator:
        An optional transformer class that take place in order to convert one or multiple column names into the performance value that is used by the rating model

        :param post_rating_transformers:
            After rating-generation, additional feature engineering can be performed.

        :param predictor:
            The object which trains and returns predictions. Defaults to LGBMClassifier

        :param estimator: Sklearn-like estimator. If predictor is set, estimator will be ignored.
         Because it sometimes can be tricky to identify the names of all the features that must be passed to predictor, the user can decide to only pass in an estimator.
         The features will then be automatically identified based on features_created from the rating_generator, post_rating_transformers and other_features.
         If predictor is set, estimator will be ignored

        :param other_features: If estimator is set and predictor is not,
        other_features allows the user to pass in additional features that are not created by the rating_generator or post_rating_transformers to the predictor.

        :param other_categorical_features: Which of the other_features are categorical.
        It is not required to duplicate the categorical features in other_features and other_categorical_features.
        Simply passing an a categorical_feature in categorical_features will add it to other_features if it doesn't already exist.

        :param train_split_date:
            Date threshold which defines which periods to train the predictor on

        :param date_column_name:
            If rating_generators are not defined and train_split_date is used, then train_column_name must be set.

        :param use_auto_create_performance_calculator:
            If true, the pre_rating_transformers will be automatically generated to ensure the performance-value is done according to good practices.
            For new users, this is recommended.

        :param column_weights:
            If auto_create_pre_transformers is True, column_weights must be set.
            It is generally used when multiple columns are used to calculate ratings and the columns need to be weighted when converting it to a performance_value.
            Even if only 1  feature is used but auto_create_pre_transformers is used,
             then it must still be created in order for auto_create_pre_transformers to know which columns needs to be transformed.

        """

        self.rating_generators: list[RatingGenerator] = rating_generators if isinstance(rating_generators, list) else [
            rating_generators]
        if rating_generators is None:
            self.rating_generators: list[RatingGenerator] = []

        self.auto_create_performance_calculator = use_auto_create_performance_calculator
        if self.auto_create_performance_calculator and not column_weights:
            raise ValueError("column_weights must be set if auto_create_pre_transformers is True")

        self.column_weights = column_weights
        if self.column_weights and isinstance(self.column_weights[0], ColumnWeight):
            self.column_weights = [self.column_weights]

        if not self.auto_create_performance_calculator and column_weights:
            logging.warning(
                "column_weights is set but auto_create_pre_transformers is False. column_weights will be ignored")

        self.performances_generator = performances_generator
        if self.auto_create_performance_calculator:
            if not self.rating_generators:
                raise ValueError("rating_generators must be set if auto_create_pre_transformers is True")
            column_names = [r.column_names for r in self.rating_generators]
            self.performances_generator = auto_create_performance_generator(column_weights=self.column_weights,
                                                                            column_names=column_names)

        self.post_rating_transformers = post_rating_transformers or []
        self.post_prediction_transformers = post_prediction_transformers or []

        self.predictor = predictor
        self.other_features = other_features or []
        self.other_categorical_features = other_categorical_features or []
        self.group_predictor_by_game_team = group_predictor_by_game_team
        self.estimator_or_transformers = estimator_or_transformers
        self.match_id_column_name = match_id_column_name
        self.team_id_column_name = team_id_column_name
        self.keep_features = keep_features

        if self.predictor is not None and estimator is not None:
            logging.warning(
                "predictor and estimator is set. estimator will be ignored. If it's intended to be used, either inject it into predictor or remove predictor")

        if self.predictor is not None and self.other_features:
            logging.warning(
                "predictor and other_features is set. other_features will be ignored. If it's intended to be used, either inject it into predictor or remove predictor")

        if self.predictor is None:
            self.predictor = create_predictor(
                rating_generators=self.rating_generators,
                other_features=self.other_features,
                other_categorical_features=self.other_categorical_features,
                post_rating_transformers=self.post_rating_transformers,
                estimator=estimator,
                group_predictor_by_game_team=self.group_predictor_by_game_team,
                match_id_column_name=self.match_id_column_name,
                team_id_column_name=self.team_id_column_name
            )

        self.predictor.set_target(PredictColumnNames.TARGET)
        self.train_split_date = train_split_date
        self.date_column_name = date_column_name
        if self.train_split_date and date_column_name is None:
            if not self.rating_generators:
                raise ValueError(
                    "date_column_name must be set if train_split_date is set and rating_generators is None")

            self.date_column_name = self.rating_generators[
                0].column_names.start_date

    def generate_historical(self, df: pd.DataFrame, matches: Optional[Union[list[Match], list[list[Match]]]] = None,
                            store_ratings: bool = True) -> pd.DataFrame:

        df = df.copy()

        if self.predictor.pred_column in df.columns:
            raise ValueError(f"Predictor column {self.predictor.pred_column} already in df columns. Remove or rename before generating predictions")

        ori_cols = df.columns.tolist()

        if matches:
            if isinstance(matches[0], Match):
                matches = [matches for _ in self.rating_generators]

        elif self.performances_generator:
            df = self.performances_generator.generate(df)

        if self.predictor.target not in df.columns:
            raise ValueError(
                f"Target {self.predictor.target} not in df columns. Target always needs to be set equal to {PredictColumnNames.TARGET}")

        for rating_idx, rating_generator in enumerate(self.rating_generators):

            rating_column_names = rating_generator.column_names

            if matches is None:
                rating_matches = convert_df_to_matches(column_names=rating_column_names, df=df,
                                                       league_identifier=LeagueIdentifier())
            else:
                rating_matches = matches[rating_idx]

            if store_ratings:
                match_ratings = rating_generator.generate_historical(matches=rating_matches, df=df)
            else:
                match_ratings = rating_generator.generate_historical(matches=rating_matches)
            for rating_feature, values in match_ratings.items():
                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature
                df[rating_feature_str] = values

        for post_rating_transformer in self.post_rating_transformers:
            df = post_rating_transformer.fit_transform(df)

        if self.date_column_name:
            if not self.train_split_date:
                self.train_split_date = df.iloc[int(len(df) / 1.3)][self.date_column_name]
            train_df = df[df[self.date_column_name] <= self.train_split_date]
        else:
            logging.warning("train date is not defined. Uses entire dataset to train predictor")
            train_df = df

        self.predictor.train(train_df)
        df = self.predictor.add_prediction(df)
        for post_prediction_transformer in self.post_prediction_transformers:
            df = post_prediction_transformer.transform(df)
        if not self.keep_features:
            df = df[ori_cols + [self.predictor.pred_column]]

        return df

    def predict(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        ori_cols = df.columns.tolist()
        for rating_idx, rating_generator in enumerate(self.rating_generators):
            rating_column_names = rating_generator.column_names

            matches = convert_df_to_matches(column_names=rating_column_names, df=df,
                                            league_identifier=LeagueIdentifier())

            match_ratings = rating_generator.generate_future(matches, df=df)
            for rating_feature in rating_generator.features_out:
                values = match_ratings[rating_feature]

                if len(self.rating_generators) > 1:
                    rating_feature_str = rating_feature + str(rating_idx)
                else:
                    rating_feature_str = rating_feature
                df[rating_feature_str] = values

        for post_rating_transformer in self.post_rating_transformers:
            df = post_rating_transformer.transform(df)

        df = self.predictor.add_prediction(df)
        for post_prediction_transformer in self.post_prediction_transformers:
            df = post_prediction_transformer.transform(df)
        if not self.keep_features:
            df = df[ori_cols + [self.predictor.pred_column]]
        return df

    @property
    def classes_(self) -> Optional[list[str]]:
        if 'classes_' not in dir(self.predictor.estimator):
            return None
        return self.predictor.estimator.classes_



from .post_transformers import LagTransformer, RollingMeanTransformer
from .pre_transformers import SkLearnTransformerWrapper, MinMaxTransformer, DiminishingValueTransformer, \
    SklearnEstimatorImputer
